###############################################################################
# Coscine (File) Explorer Example Application
# Copyright (c) 2018-2021 RWTH Aachen University
###############################################################################
# This python program implements a simple file explorer for Coscine.
# It uses the Coscine API through the Coscine Python Client.
# The purpose of this program is to provide a simple and extensible platform
# showcasing how one can integrate Coscine functionality into (graphical)
# applications.
# Visit https://coscine.rwth-aachen.de for more information about Coscine.
###############################################################################

import sys
import os
import json
from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import coscine

###############################################################################

__version__ = coscine.about.__version__

###############################################################################

WIDTH = 1024
HEIGHT = 640

###############################################################################

class Explorer(QMainWindow):

	def __init__(self):
		super().__init__()
		self.setWindowTitle("Coscine Explorer " + __version__)
		self.create_menubar()
		self.load_data()
		self.threadpool = QThreadPool()
		self.coscine = coscine.Client(
			self.config["token"],
			lang = self.config["lang"]
		)
		self.setMinimumSize(WIDTH, HEIGHT)
		self.setWindowIcon(self.winicon)
		self.setCentralWidget(GUI(self))
		self.show()

###############################################################################

	def create_menubar(self):
		authAction = QAction("Authenticate", self)
		authAction.setShortcut("Ctrl+A")
		authAction.triggered.connect(self.authenticate)

		langAction = QAction("Language", self)
		langAction.triggered.connect(self.set_language)

		quitAction = QAction("Quit", self)
		quitAction.setShortcut("Ctrl+Q")
		quitAction.triggered.connect(self.quit)

		menubar = self.menuBar()
		menu = menubar.addMenu("Menu")
		menu.addAction(authAction)
		menu.addAction(langAction)
		menu.addAction(quitAction)

###############################################################################

	def quit(self):
		qApp.quit()

###############################################################################

	def load_icons(self):
		self.fileicons = {}
		path = "data/file-icons/16px/"
		files = os.listdir(path)
		for file in files:
			name = os.path.splitext(file)[0]
			self.fileicons[name] = QIcon(path + file)

###############################################################################

	def load_config(self):
		fd = open("config.json", "rt")
		data = fd.read()
		fd.close()
		self.config = json.loads(data)

###############################################################################

	def load_data(self):
		self.load_config()
		self.load_icons()
		self.winicon = QIcon("data/favicon.ico")
		self.logo = QPixmap("data/logo.png")

###############################################################################

	def authenticate(self):
		text, okPressed = QInputDialog.getText(self,
			"Authenticate", "Access Token:", QLineEdit.Normal, ""
		)
		if okPressed and text != "":
			print(text)

###############################################################################

	def set_language(self):
		items = ("Deutsch", "English")
		item, okPressed = QInputDialog.getItem(self,
			"Set Language", "Language:", items, 0, False
		)
		if okPressed and item:
			print(item)

###############################################################################

class WorkerSignals(QObject):
	finished = pyqtSignal(object)

###############################################################################

class WorkerThread(QRunnable):

	def __init__(self, fn, *args, **kwargs):
		super(WorkerThread, self).__init__()
		self.fn = fn
		self.args = args
		self.kwargs = kwargs
		self.signals = WorkerSignals()

###############################################################################

	def run(self):
		result = self.fn(*self.args, **self.kwargs)
		self.signals.finished.emit(result)

###############################################################################

class GUI(QWidget):

	def __init__(self, app):
		super().__init__()

		app.sideBar = SideBar(app)
		app.fileManager = FileManager(app)

		hbox = QHBoxLayout()
		hbox.addWidget(app.sideBar)
		hbox.addWidget(app.fileManager, 1)
		self.setLayout(hbox)

###############################################################################

class SideBar(QWidget):

	def __init__(self, app):
		super().__init__()

		labels = (QLabel("Projects"), QLabel("Resources"))
		logo = QLabel()
		logo.setPixmap(app.logo)

		self.projects = QComboBox()
		self.projects.currentIndexChanged.connect(self.switch_project)

		self.resources = QComboBox()
		self.resources.currentIndexChanged.connect(self.switch_resource)

		worker = WorkerThread(app.coscine.projects)
		worker.signals.finished.connect(self.update_projects)
		app.threadpool.start(worker)

		logbox = QPlainTextEdit()

		vbox = QVBoxLayout()
		vbox.addWidget(logo)
		vbox.addWidget(labels[0])
		vbox.addWidget(self.projects)
		vbox.addWidget(labels[1])
		vbox.addWidget(self.resources)
		vbox.addWidget(logbox)
		self.setLayout(vbox)
		self.app = app

###############################################################################

	def update_projects(self, projects):
		for project in projects:
			self.projects.addItem(self.app.winicon,	project.name, project)

###############################################################################

	def fetch_project(self, index):
		project = self.projects.currentData()
		worker = WorkerThread(project.resources)
		worker.signals.finished.connect(self.update_resources)
		self.app.threadpool.start(worker)

###############################################################################

	def update_resources(self, resources):
		self.resources.clear()
		self.app.fileManager.filetree.clear()
		for resource in resources:
			self.resources.addItem(self.app.winicon, resource.name, resource)

###############################################################################

	def switch_project(self, index):
		self.fetch_project(index)

###############################################################################

	def switch_resource(self, index):
		resource = self.resources.currentData()
		if resource:
			self.app.fileManager.fetch_files(resource)

###############################################################################

class FileManager(QWidget):

	def __init__(self, app):
		super().__init__()

		label = QLabel("File Explorer")
		searchBar = QLineEdit()
		searchBar.setPlaceholderText("Search")
		searchBar.textChanged.connect(self.search)

		filetree = QTreeWidget()
		headers = ["Name", "Type", "Size"]
		filetree.setSelectionMode(QAbstractItemView.SingleSelection)
		filetree.setContextMenuPolicy(Qt.CustomContextMenu)
		filetree.customContextMenuRequested.connect(self.contextMenu)
		filetree.setAlternatingRowColors(True)
		filetree.setHeaderLabels(headers)
		self.filetree = filetree

		uploadButton = QPushButton("Upload")
		uploadButton.clicked.connect(self.upload)

		downloadButton = QPushButton("Download")
		downloadButton.clicked.connect(self.download)

		hbox = QHBoxLayout()
		hbox.addWidget(uploadButton)
		hbox.addWidget(downloadButton)

		progressBar = QProgressBar()

		vbox = QVBoxLayout()
		vbox.addWidget(searchBar)
		vbox.addWidget(label)
		vbox.addWidget(filetree)
		vbox.addLayout(hbox)
		vbox.addWidget(progressBar)
		self.setLayout(vbox)
		self.setAcceptDrops(True)
		self.app = app

###############################################################################

	def fetch_files(self, resource):
		self.filetree.clear()
		files = resource.objects()
		self.files = files
		for f in files:
			columns = [f.name, f.type, str(f.size)]
			item = QTreeWidgetItem(self.filetree, columns)
			if f.type in self.app.fileicons:
				item.setIcon(0, self.app.fileicons[f.type])
			else:
				item.setIcon(0, self.app.fileicons["_blank"])
		self.filetree.resizeColumnToContents(0)

###############################################################################

	def search(self, text):
		iterator = QTreeWidgetItemIterator(self.filetree)
		while iterator.value():
			item = iterator.value()
			if item.childCount() == 0:
				item.setHidden(True)
			iterator += 1
		items = self.filetree.findItems(text, Qt.MatchContains | Qt.MatchRecursive)
		for item in items:
			item.setHidden(False)

###############################################################################

	def contextMenu(self, position):
		indexes = self.filetree.selectedIndexes()
		if not indexes:
			return

		for ix in indexes:
			print(ix.data())

		filepath = indexes[0].data()

		metadataAction = QAction("Edit Metadata")
		metadataAction.triggered.connect(self.edit_metadata)
		
		downloadAction = QAction("Download")
		downloadAction.triggered.connect(self.download)
		
		menu = QMenu()
		menu.addAction(downloadAction)
		menu.addAction(metadataAction)
		menu.addAction("Delete")
		menu.exec_(self.filetree.viewport().mapToGlobal(position))

###############################################################################

	def upload(self):
		files, _ = QFileDialog.getOpenFileNames()
		if files:
			for file in files:
				print(file)
				#self.app.coscine.upload_file(file)

###############################################################################

	def download(self):
		indexes = self.filetree.selectedIndexes()
		if not indexes:
			msg = QMessageBox().warning(
				self,
				"Coscine Explorer - Error",
				"No file selected!"
			)
			msg.exec_()
			return

		for ix in indexes:
			print(ix.data())

		filename = indexes[0].data()
		resource = self.app.sideBar.resources.currentData()

		path, _ = QFileDialog.getSaveFileName()
		if path:
			self.app.coscine.download_file(resource, filename, path)

###############################################################################

	def edit_metadata(self):
		indexes = self.filetree.selectedIndexes()
		if not indexes:
			msg = QMessageBox().warning(
				self,
				"Coscine Explorer - Error",
				"No file selected!"
			)
			msg.exec_()
			return

		filename = indexes[0].data()
		row = indexes[0].row()
		handle = self.files[row]
		editor = MetadataEditor(self.app, edit=handle)

###############################################################################

class MetadataEditor(QDialog):

	def __init__(self, app, edit=None):
		super().__init__()
		self.app = app
		self.edit = edit
		self.createFormGroupBox()
		self.setWindowTitle("Metadata Editor")
		buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
		buttonBox.accepted.connect(self.accept)
		buttonBox.rejected.connect(self.reject)

		scrollArea = QScrollArea()
		scrollArea.setWidget(self.formGroupBox)
		scrollArea.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

		mainLayout = QVBoxLayout()
		mainLayout.addWidget(scrollArea)
		mainLayout.addWidget(buttonBox)
		self.setLayout(mainLayout)
		#self.show()
		self.exec_()

###############################################################################

	def createFormGroupBox(self):
		self.formGroupBox = QGroupBox("Metadata")
		form = self.edit.MetadataForm()
		layout = QFormLayout()
		print(form)
		for key in form.keys():
			if form.is_controlled(key):
				element = QComboBox()
				element.setEditable(True)
				element.setInsertPolicy(QComboBox.NoInsert)
				element.completer().setCompletionMode(QCompleter.PopupCompletion)
				vocabulary = form.get_vocabulary(key)
				for entry in vocabulary:
					element.addItem(entry)
			else:
				element = QLineEdit(form[key])
			layout.addRow(QLabel(key), element)
		self.formGroupBox.setLayout(layout)

###############################################################################